# src/pyneurorg/core/neuron_models.py

"""
Collection of predefined neuron models for use with Brian2 in pyneurorg simulations.
(Resto da docstring como antes)
"""

import brian2 as b2
from brian2.units.fundamentalunits import DIMENSIONLESS
import numpy as np

def _add_common_stim_syn_vars_and_flags(base_eqs_str: str,
                                        base_namespace_dict: dict,
                                        num_stim_flags: int = 16,
                                        current_vars_units: str = "amp"):
    current_declarations = [
        f"I_synaptic : {current_vars_units} # Sum of synaptic currents (e.g., from non-conductance based synapses)",
        f"I_stimulus_sum : {current_vars_units} # Sum of external timed stimuli"
    ]

    flag_eqs_list = []
    default_flags_init_ns = {}
    for i in range(num_stim_flags):
        flag_name = f"stf{i}" # Stimulus Target Flag
        flag_eqs_list.append(f"{flag_name} : boolean")
        default_flags_init_ns[f"{flag_name}_default_init"] = False

    full_eqs_str_parts = []
    full_eqs_str_parts.extend(current_declarations)
    if flag_eqs_list:
        full_eqs_str_parts.extend(flag_eqs_list)
    full_eqs_str_parts.append(base_eqs_str)

    full_eqs_str = "\n".join(filter(None, full_eqs_str_parts))

    updated_namespace = base_namespace_dict.copy()

    default_current_unit_obj = b2.amp
    if current_vars_units == "amp/meter**2":
        default_current_unit_obj = b2.amp / b2.meter**2
    elif current_vars_units == "1": # Dimensionless
        default_current_unit_obj = b2.Quantity(1, dim=DIMENSIONLESS)


    updated_namespace.setdefault('I_synaptic_default_init', 0 * default_current_unit_obj)
    updated_namespace.setdefault('I_stimulus_sum_default_init', 0 * default_current_unit_obj)
    updated_namespace.update(default_flags_init_ns)

    return full_eqs_str, updated_namespace


def LIFNeuron(tau_m=10*b2.ms, v_rest=0*b2.mV, v_reset=0*b2.mV,
              v_thresh=20*b2.mV, R_m=100*b2.Mohm, I_tonic=0*b2.nA,
              refractory_period=2*b2.ms, num_stim_flags=16):
    base_eqs = """
    dv/dt = (-(v - v_rest_val) + R_m_val * (I_synaptic + I_stimulus_sum + I_tonic_val)) / tau_m_val : volt (unless refractory)
    """
    base_namespace = {
        'tau_m_val': tau_m, 'v_rest_val': v_rest, 'R_m_val': R_m,
        'I_tonic_val': I_tonic, 'v_default_init': v_rest,
    }
    full_eqs, full_namespace = _add_common_stim_syn_vars_and_flags(
        base_eqs, base_namespace, num_stim_flags, current_vars_units="amp"
    )
    return {
        'model': full_eqs, 'threshold': f'v > {v_thresh!r}', 'reset': f'v = {v_reset!r}',
        'refractory': refractory_period, 'namespace': full_namespace, 'method': 'exact'
    }

def ConductanceLIFNeuron(tau_m=20*b2.ms, v_rest=-65*b2.mV, v_reset=-65*b2.mV,
                         v_thresh=-50*b2.mV, R_m=100*b2.Mohm, I_tonic=0*b2.nA,
                         refractory_period=3*b2.ms,
                         E_exc=0*b2.mV, E_inh=-80*b2.mV,
                         tau_g_exc=5*b2.ms, tau_g_inh=10*b2.ms,
                         num_stim_flags=16):
    base_eqs = """
    I_conductance = g_exc*(E_exc_val - v) + g_inh*(E_inh_val - v) : amp
    dv/dt = (-(v - v_rest_val) + R_m_val * (I_conductance + I_synaptic + I_stimulus_sum + I_tonic_val)) / tau_m_val : volt (unless refractory)
    dg_exc/dt = -g_exc / tau_g_exc_val : siemens (unless refractory)
    dg_inh/dt = -g_inh / tau_g_inh_val : siemens (unless refractory)
    """
    base_namespace = {
        'tau_m_val': tau_m, 'v_rest_val': v_rest, 'R_m_val': R_m,
        'I_tonic_val': I_tonic, 'E_exc_val': E_exc, 'E_inh_val': E_inh,
        'tau_g_exc_val': tau_g_exc, 'tau_g_inh_val': tau_g_inh,
        'v_default_init': v_rest,
        'g_exc_default_init': 0*b2.nS,
        'g_inh_default_init': 0*b2.nS
    }
    full_eqs, full_namespace = _add_common_stim_syn_vars_and_flags(
        base_eqs, base_namespace, num_stim_flags, current_vars_units="amp"
    )
    return {
        'model': full_eqs, 'threshold': f'v > {v_thresh!r}', 'reset': f'v = {v_reset!r}',
        'refractory': refractory_period, 'namespace': full_namespace,
        'method': 'euler'
    }


def AdExNeuron(C_m=281*b2.pF, g_L=30*b2.nS, E_L=-70.6*b2.mV,
               V_T=-50.4*b2.mV, Delta_T=2*b2.mV, tau_w=144*b2.ms,
               a=4*b2.nS, adex_b_param=0.0805*b2.nA, V_reset=-70.6*b2.mV,
               V_peak=0*b2.mV, I_tonic=0*b2.nA, refractory_period=0*b2.ms,
               num_stim_flags=16):
    base_eqs = """
    dv/dt = (g_L_val * (E_L_val - v) + g_L_val * Delta_T_val * exp((v - V_T_val)/Delta_T_val) - w + I_synaptic + I_stimulus_sum + I_tonic_val) / C_m_val : volt (unless refractory)
    dw/dt = (a_val * (v - E_L_val) - w) / tau_w_val : amp (unless refractory)
    """
    temp_base_namespace = {
        'C_m_val': C_m, 'g_L_val': g_L, 'E_L_val': E_L, 'V_T_val': V_T, 'Delta_T_val': Delta_T,
        'tau_w_val': tau_w, 'a_val': a, 'I_tonic_val': I_tonic,
        'v_default_init': E_L, 'w_default_init': 0*b2.nA,
    }
    full_eqs, full_namespace = _add_common_stim_syn_vars_and_flags(
        base_eqs, temp_base_namespace, num_stim_flags, current_vars_units="amp"
    )
    return {
        'model': full_eqs,
        'threshold': f'v > {V_peak!r}',
        'reset': f'v = {V_reset!r}; w += {adex_b_param!r}',
        'refractory': refractory_period,
        'namespace': full_namespace,
        'method': 'euler' 
    }

def SimpleEIFNeuron(C_m=281*b2.pF, g_L=30*b2.nS, E_L=-70.6*b2.mV,
                    V_T=-50.4*b2.mV, Delta_T=2*b2.mV, V_reset=-70.6*b2.mV,
                    V_peak=0*b2.mV, I_tonic=0*b2.nA, refractory_period=2*b2.ms,
                    num_stim_flags=16):
    """
    Simple Exponential Integrate-and-Fire (EIF) neuron model.
    This is the AdEx model without the adaptation variable 'w'.
    """
    base_eqs = """
    dv/dt = (g_L_val * (E_L_val - v) + g_L_val * Delta_T_val * exp((v - V_T_val)/Delta_T_val) + I_synaptic + I_stimulus_sum + I_tonic_val) / C_m_val : volt (unless refractory)
    """
    temp_base_namespace = {
        'C_m_val': C_m, 'g_L_val': g_L, 'E_L_val': E_L, 'V_T_val': V_T, 'Delta_T_val': Delta_T,
        'I_tonic_val': I_tonic,
        'v_default_init': E_L,
    }
    full_eqs, full_namespace = _add_common_stim_syn_vars_and_flags(
        base_eqs, temp_base_namespace, num_stim_flags, current_vars_units="amp"
    )
    return {
        'model': full_eqs,
        'threshold': f'v > {V_peak!r}',
        'reset': f'v = {V_reset!r}', # No w to update
        'refractory': refractory_period,
        'namespace': full_namespace,
        'method': 'euler' # Euler or exponential_euler
    }


def SimpleHHNeuron(C_m=1*b2.uF/b2.cm**2, E_Na=50*b2.mV, E_K=-77*b2.mV, E_L=-54.4*b2.mV,
                   g_Na_bar=120*b2.mS/b2.cm**2, g_K_bar=36*b2.mS/b2.cm**2, g_L_bar=0.3*b2.mS/b2.cm**2,
                   V_T_hh=-60*b2.mV, I_tonic=0*b2.uA/b2.cm**2, refractory_period=0*b2.ms,
                   num_stim_flags=16, area=1000*b2.um**2):
    numerical_spike_threshold = 0*b2.mV
    base_eqs = """
    dv/dt = (I_Na + I_K + I_L + I_synaptic + I_stimulus_sum + I_tonic_val) / C_m_val : volt (unless refractory)
    I_Na = g_Na_bar_val * m**3 * h * (E_Na_val - v) : amp/meter**2
    dm/dt = alpha_m * (1-m) - beta_m * m : 1
    dh/dt = alpha_h * (1-h) - beta_h * h : 1
    alpha_m = (0.1/mV) * (v - (V_T_hh_val + 25*mV)) / (1 - exp(-(v - (V_T_hh_val + 25*mV))/(10*mV))) / ms : Hz
    beta_m = (4.0) * exp(-(v - (V_T_hh_val + 0*mV))/(18*mV)) / ms : Hz
    alpha_h = (0.07) * exp(-(v - (V_T_hh_val + 0*mV))/(20*mV)) / ms : Hz
    beta_h = (1.0) / (1 + exp(-(v - (V_T_hh_val + 30*mV))/(10*mV))) / ms : Hz
    I_K = g_K_bar_val * n**4 * (E_K_val - v) : amp/meter**2
    dn/dt = alpha_n * (1-n) - beta_n * n : 1
    alpha_n = (0.01/mV) * (v - (V_T_hh_val + 10*mV)) / (1 - exp(-(v - (V_T_hh_val + 10*mV))/(10*mV))) / ms : Hz
    beta_n = (0.125) * exp(-(v - (V_T_hh_val + 0*mV))/(80*mV)) / ms : Hz
    I_L = g_L_bar_val * (E_L_val - v) : amp/meter**2
    area_val : meter**2 
    """
    base_namespace = {
        'C_m_val': C_m, 'E_Na_val': E_Na, 'E_K_val': E_K, 'E_L_val': E_L,
        'g_Na_bar_val': g_Na_bar, 'g_K_bar_val': g_K_bar, 'g_L_bar_val': g_L_bar,
        'V_T_hh_val': V_T_hh, 'I_tonic_val': I_tonic,
        'v_default_init': E_L,
        'm_default_init': 0.0529,
        'h_default_init': 0.5961,
        'n_default_init': 0.3177,
        'area_val_default_init': area 
    }

    current_density_units_str = "amp/meter**2"
    full_eqs, full_namespace = _add_common_stim_syn_vars_and_flags(
        base_eqs, base_namespace, num_stim_flags,
        current_vars_units=current_density_units_str
    )
    if 'area_val' not in full_namespace:
         full_namespace['area_val'] = area 
    if not isinstance(full_namespace['area_val'], b2.Quantity):
        if isinstance(full_namespace['area_val'], (int, float)): #pragma: no cover
            full_namespace['area_val'] = full_namespace['area_val'] * b2.meter**2
        else: 
            try: #pragma: no cover
                full_namespace['area_val'] = full_namespace['area_val'].in_units(b2.meter**2)
            except: #pragma: no cover
                raise TypeError("area_val for SimpleHHNeuron must be convertible to meter**2")


    return {
        'model': full_eqs,
        'threshold': f'v > {numerical_spike_threshold!r}',
        'reset': '',
        'refractory': refractory_period,
        'namespace': full_namespace,
        'method': 'exponential_euler'
    }

def LIFCalciumFluorNeuron(tau_m=10*b2.ms, v_rest=0*b2.mV, v_reset=0*b2.mV,
                          v_thresh=20*b2.mV, R_m=100*b2.Mohm, I_tonic=0*b2.nA,
                          refractory_period=2*b2.ms,
                          tau_ca=50*b2.ms, ca_spike_increment=0.2,
                          tau_f=100*b2.ms, k_f=0.5,
                          num_stim_flags=16):
    base_eqs = """
    dv/dt = (-(v - v_rest_val) + R_m_val * (I_synaptic + I_stimulus_sum + I_tonic_val)) / tau_m_val : volt (unless refractory)
    dCa/dt = -Ca / tau_ca_val : 1 (unless refractory)
    dF/dt = (k_f_val * Ca - F) / tau_f_val : 1 (unless refractory)
    """
    temp_base_namespace = { 
        'tau_m_val': tau_m, 'v_rest_val': v_rest, 'R_m_val': R_m,
        'I_tonic_val': I_tonic, 
        'tau_ca_val': tau_ca, 'tau_f_val': tau_f, 'k_f_val': k_f,
        'v_default_init': v_rest,
        'Ca_default_init': 0.0,
        'F_default_init': 0.0,
    }
    full_eqs, full_namespace = _add_common_stim_syn_vars_and_flags(
        base_eqs, temp_base_namespace, num_stim_flags, current_vars_units="amp"
    )
    full_namespace['ca_spike_increment_val'] = ca_spike_increment

    return {
        'model': full_eqs,
        'threshold': f'v > {v_thresh!r}',
        'reset': f'v = {v_reset!r}; Ca += ca_spike_increment_val',
        'refractory': refractory_period,
        'namespace': full_namespace,
        'method': 'euler'
    }


# src/pyneurorg/core/synapse_models.py

"""
Collection of predefined synapse models for use with Brian2 in pyneurorg simulations.

These models typically describe how a presynaptic spike influences a postsynaptic neuron,
often by modulating a conductance variable (e.g., 'g_exc', 'g_inh', 'g_nmda')
in the postsynaptic neuron's model. The decay dynamics of these conductances
are generally assumed to be handled by the postsynaptic neuron model itself,
unless specified otherwise by a particular synapse model that defines its own
internal state variables for conductance.
"""

import brian2 as b2

def StaticConductionSynapse(weight_increment=1.0*b2.nS, target_conductance_var='g_exc'):
    """
    A simple synapse that causes an instantaneous, fixed increase in the
    `target_conductance_var` of the postsynaptic neuron upon a presynaptic spike.

    The decay of this conductance is assumed to be handled by the postsynaptic
    neuron model (e.g., if it has an equation like `dg_exc/dt = -g_exc/tau_exc`).

    Parameters
    ----------
    weight_increment : brian2.units.Quantity, optional
        The amount of conductance to add (default: 1.0 nS).
    target_conductance_var : str, optional
        The name of the conductance variable in the postsynaptic neuron
        (e.g., 'g_exc', 'g_ampa').
    """
    model_eqs = """
    w_inc : siemens (constant) # Weight increment per spike
    """
    on_pre_action = f'{target_conductance_var}_post += w_inc'

    return {
        'model': model_eqs,
        'on_pre': on_pre_action,
        'namespace': {
            # w_inc is a per-synapse parameter.
            # The `weight_increment` arg can be used as a default when creating connections,
            # which will be assigned to S.w_inc by the Organoid/Simulator.
        }
    }

def ExponentialIncrementSynapse(weight_increment=0.5*b2.nS, target_conductance_var='g_exc'):
    """
    Synapse that causes an instantaneous increase in the `target_conductance_var`
    of the postsynaptic neuron.

    The name implies an expectation of exponential decay for this conductance,
    but the decay dynamics *must* be implemented in the postsynaptic neuron model.
    This synapse model itself does not implement decay.

    Parameters
    ----------
    weight_increment : brian2.units.Quantity, optional
        The amount of conductance to add (default: 0.5 nS).
    target_conductance_var : str, optional
        The name of the conductance variable in the postsynaptic neuron.
    """
    # Functionally identical to StaticConductionSynapse in its implementation,
    # the name difference is conceptual for the user.
    model_eqs = """
    w_inc : siemens (constant) # Weight increment per spike
    """
    on_pre_action = f'{target_conductance_var}_post += w_inc'

    return {
        'model': model_eqs,
        'on_pre': on_pre_action,
        'namespace': {}
    }

def STPSynapse(U_stp=0.5, tau_facilitation=50*b2.ms, tau_depression=200*b2.ms,
               base_weight=1.0*b2.nS, target_conductance_var='g_exc'):
    """
    Conductance-based synapse with short-term plasticity (Tsodyks-Markram model).
    The synaptic efficacy changes based on recent presynaptic activity.
    The actual conductance decay is handled by the postsynaptic neuron.

    Parameters
    ----------
    U_stp : float, optional
        Utilization of synaptic efficacy (fraction of resources used by a spike, unitless, default: 0.5).
    tau_facilitation : brian2.units.Quantity, optional
        Time constant for recovery of facilitation variable `u_stp` (default: 50 ms).
    tau_depression : brian2.units.Quantity, optional
        Time constant for recovery of depression variable `x_stp` (default: 200 ms).
    base_weight : brian2.units.Quantity, optional
        The baseline synaptic weight/conductance increment when u_stp=U_stp and x_stp=1 (default: 1.0 nS).
    target_conductance_var : str, optional
        The name of the conductance variable in the postsynaptic neuron.
    """
    model_eqs = """
    dx_stp/dt = (1 - x_stp) / tau_d_val : 1 (clock-driven) # Recovery of available resources
    du_stp/dt = -u_stp / tau_f_val : 1 (clock-driven)     # Recovery of facilitation factor
    
    # Per-synapse parameters (can be made heterogeneous)
    U_stp_val : 1 (constant)
    tau_f_val : second (constant)
    tau_d_val : second (constant)
    w_base : siemens (constant) # Baseline weight
    """
    # On a presynaptic spike:
    # 1. Conductance change: g_post += w_base * u_eff * x_eff
    #    (u_eff and x_eff are values of u_stp and x_stp *before* this spike's update)
    # 2. Update u_stp (facilitation): u_stp_new = u_stp + U_stp_val * (1 - u_stp)
    # 3. Update x_stp (depression): x_stp_new = x_stp * (1 - u_eff_for_this_spike)
    #    where u_eff_for_this_spike is the (updated) u_stp
    #
    # Brian2's on_pre executes statements sequentially.
    # We need to use u_stp and x_stp *before* they are updated for the current spike's effect.
    # A common way is to calculate the effective u for this spike, then apply conductance, then update x.
    
    on_pre_action = f"""
    u_eff_stp = u_stp_val + U_stp_val * (1 - u_stp_val) # Effective u for this spike (facilitated)
    conductance_change = w_base * u_eff_stp * x_stp_val # Calculate conductance change
    {target_conductance_var}_post += conductance_change  # Apply to postsynaptic neuron

    u_stp_val = u_eff_stp             # Update u_stp for the next spike
    x_stp_val = x_stp_val * (1 - u_eff_stp) # Update x_stp (depression based on the used u_eff_stp)
    """
    # Note: The above formulation for u_eff_stp might be slightly different from some Tsodyks-Markram versions.
    # A common version:
    # on_pre:
    #   g_post += A * u * x
    #   u = u + U * (1-u)
    #   x = x * (1-u_prev_spike) where u_prev_spike is u *before* U is added.
    # Let's use a formulation closer to Brian2 examples:
    on_pre_action_v2 = f"""
    effective_u = u_stp + U_stp_val * (1 - u_stp)
    {target_conductance_var}_post += w_base * effective_u * x_stp
    u_stp = effective_u
    x_stp = x_stp * (1 - effective_u)
    """
    # Initial values for u_stp and x_stp are important.
    # u_stp usually starts at 0 or U_stp, x_stp at 1.

    return {
        'model': model_eqs,
        'on_pre': on_pre_action_v2, # Use v2 which is more standard
        'namespace': {
            'U_stp_val_default_init': U_stp, # Default for the parameter U_stp_val if not set
            'tau_f_val_default_init': tau_facilitation,
            'tau_d_val_default_init': tau_depression,
            'w_base_default_init': base_weight,
            'x_stp_default_init': 1.0, # Start with full resources
            'u_stp_default_init': 0.0  # Start with no facilitation (will jump to U_stp on first spike)
                                      # or U_stp if u_eff_stp is u_stp_val in the on_pre
                                      # If on_pre is u_stp = u_stp + U(1-u_stp), then u_stp should start at 0.
                                      # If on_pre is u_stp = U + u_stp*(1-U), then u_stp should start at U.
                                      # The v2 uses u_stp_default_init = 0.0 and first spike u_eff = U.
        },
        'method': 'exact' # ODEs for x_stp and u_stp are linear
    }

def NMDASynapse(weight_nmda=0.5*b2.nS, target_conductance_var='g_nmda',
                mg_conc=1.0, eta_nmda=0.08, gamma_nmda=0.062):
    """
    NMDA receptor-mediated conductance synapse with voltage-dependent magnesium block.
    The actual conductance decay (e.g., for g_nmda) is handled by the postsynaptic neuron.
    This synapse model calculates the effective current based on g_nmda and voltage block.

    This model assumes the postsynaptic neuron has 'g_nmda' which this synapse increments,
    AND the postsynaptic neuron's dv/dt includes a term like:
    `I_nmda_current = g_nmda * (1.0 / (1.0 + eta_nmda_val * mg_conc_val * exp(-gamma_nmda_val * v_post/mV))) * (E_nmda - v_post)`
    OR, this synapse can directly write to a current I_nmda_summed : amp.

    Let's choose the simpler path: this synapse increments target_conductance_var,
    and the postsynaptic neuron model *must* implement the B(V) term in its equations.

    Parameters
    ----------
    weight_nmda : brian2.units.Quantity, optional
        The amount of g_nmda conductance to add on a spike (default: 0.5 nS).
    target_conductance_var : str, optional
        Name of the NMDA conductance variable in the postsynaptic neuron (e.g., 'g_nmda').
    mg_conc, eta_nmda, gamma_nmda : float, optional
        Parameters for the magnesium block function B(V). These would typically be
        part of the postsynaptic neuron's namespace if B(V) is computed there.
        If this synapse were to compute the current directly, they'd be here.
        Since we are only incrementing a conductance, these are not directly used
        by *this synapse model's equations or on_pre*, but are listed for conceptual completeness
        as they are essential for the NMDA current calculation in the postsynaptic neuron.
    """
    model_eqs = """
    w_nmda_inc : siemens (constant) # NMDA conductance increment per spike
    """
    on_pre_action = f'{target_conductance_var}_post += w_nmda_inc'

    # The namespace for this synapse model itself is minimal if it only increments.
    # The parameters mg_conc, eta_nmda, gamma_nmda would be needed by the
    # postsynaptic neuron model.
    return {
        'model': model_eqs,
        'on_pre': on_pre_action,
        'namespace': {}
    }

def GABAAIncrementSynapse(weight_gabaa=1.0*b2.nS, target_conductance_var='g_gabaa'):
    """
    GABA_A receptor-mediated inhibitory synapse (fast inhibition).
    Increments `target_conductance_var` (e.g., 'g_gabaa') in the postsynaptic neuron.
    The decay of this conductance is handled by the postsynaptic neuron.

    Parameters
    ----------
    weight_gabaa : brian2.units.Quantity, optional
        Amount of g_gabaa conductance to add on a spike (default: 1.0 nS).
    target_conductance_var : str, optional
        Name of the GABA_A conductance variable (e.g., 'g_gabaa', 'g_inh').
    """
    model_eqs = """
    w_gabaa_inc : siemens (constant)
    """
    on_pre_action = f'{target_conductance_var}_post += w_gabaa_inc'
    return {
        'model': model_eqs,
        'on_pre': on_pre_action,
        'namespace': {}
    }

def GABABIncrementSynapse(weight_gabab=0.2*b2.nS, target_conductance_var='g_gabab'):
    """
    Simplified GABA_B receptor-mediated inhibitory synapse (slower inhibition).
    Increments `target_conductance_var` (e.g., 'g_gabab') in the postsynaptic neuron.
    The postsynaptic neuron is assumed to have a slower decay time constant for this
    conductance compared to GABA_A.

    Parameters
    ----------
    weight_gabab : brian2.units.Quantity, optional
        Amount of g_gabab conductance to add on a spike (default: 0.2 nS).
    target_conductance_var : str, optional
        Name of the GABA_B conductance variable (e.g., 'g_gabab').
    """
    model_eqs = """
    w_gabab_inc : siemens (constant)
    """
    on_pre_action = f'{target_conductance_var}_post += w_gabab_inc'
    return {
        'model': model_eqs,
        'on_pre': on_pre_action,
        'namespace': {}
    }


def GapJunctionSynapse(gap_conductance=0.1*b2.nS):
    """
    Electrical synapse (gap junction).
    Allows direct flow of current between connected neurons proportional to
    their voltage difference and `gap_conductance`.

    The current `I_gap_post = g_gap * (v_pre - v_post)` is injected into the
    postsynaptic neuron. The postsynaptic neuron model should include a variable
    like `I_gap_summed : amp` in its equations, which this synapse will target.
    The `(summed)` flag for `I_gap_summed` in the neuron handles multiple gap junctions.

    Parameters
    ----------
    gap_conductance : brian2.units.Quantity, optional
        Conductance of the gap junction (default: 0.1 nS).
    """
    # This model defines a current that is continuously active, not just on_pre.
    # Brian2 handles this by defining the equation that links pre and post.
    # The variable I_gap_summed_post will be summed into the postsynaptic neuron.
    # The postsynaptic neuron must have a variable I_gap_summed : amp.
    model_eqs = """
    g_gap : siemens (constant) # Gap junction conductance
    I_gap_summed_post = g_gap * (v_pre - v_post) : amp (summed)
    """
    # No on_pre or on_post needed as current flows based on voltage difference.
    return {
        'model': model_eqs,
        'on_pre': '', # No action on presynaptic spike per se for current flow
        'namespace': {
            # g_gap is a per-synapse parameter.
            # The `gap_conductance` arg can be used as default.
        }
    }


# pyneurorg/visualization/spike_plotter.py

"""
Functions for visualizing spike train data and membrane potential traces
from pyneurorg simulations, using Matplotlib.
"""

import matplotlib.pyplot as plt
import numpy as np
import brian2 as b2

def plot_raster(spike_indices, spike_times, duration=None, ax=None,
                marker_size=2, marker_color='black', title="Raster Plot",
                time_unit_display=b2.ms, ylabel="Neuron Index"):
    """
    Generates a raster plot of spike activity.

    Parameters
    ----------
    spike_indices : array-like or brian2.core.variables.VariableView
        Array of neuron indices corresponding to each spike. Typically `SpikeMonitor.i`.
    spike_times : brian2.core.variables.VariableView or brian2.units.fundamentalunits.Quantity
        Array of spike times. Typically `SpikeMonitor.t`. Will be converted to `time_unit_display`.
    duration : brian2.units.fundamentalunits.Quantity or float, optional
        Total duration for the x-axis limit. If a Brian2 Quantity, it's converted
        to `time_unit_display`. If a float, it's assumed to be in `time_unit_display`.
    ax : matplotlib.axes.Axes, optional
        An existing Matplotlib Axes object to plot on.
    marker_size : int, optional
        Size of the markers for spikes.
    marker_color : str, optional
        Color of the spike markers.
    title : str, optional
        Title for the plot.
    time_unit_display : brian2.units.fundamentalunits.UnitSymbol, optional
        The Brian2 unit to use for displaying time on the x-axis (default: b2.ms).
    ylabel : str, optional
        Label for the y-axis.
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=(10, 4))

    # --- CORREÇÃO PARA LIDAR COM VariableView E Quantity ---
    # Get numerical values of spike_times in the desired display unit
    try:
        # For VariableView or Quantity, dividing by the unit gives a plain NumPy array of numbers
        plot_times_val = np.asarray(spike_times / time_unit_display)
    except AttributeError: # If spike_times doesn't have units (e.g., already a raw array in correct scale)
        plot_times_val = np.asarray(spike_times)
    except Exception as e:
        raise TypeError(f"Could not process spike_times. Expected Brian2 VariableView/Quantity or array-like. Error: {e}")

    # Ensure spike_indices is a NumPy array
    plot_indices_val = np.asarray(spike_indices)
    if not (plot_indices_val.ndim == 1 and plot_times_val.ndim == 1 and len(plot_indices_val) == len(plot_times_val)):
        if len(plot_indices_val) == 0 and len(plot_times_val) == 0: # Empty, which is fine
            pass
        else:
            raise ValueError("spike_indices and spike_times must be 1D arrays of the same length.")


    plot_duration_val = None
    if duration is not None:
        if isinstance(duration, b2.Quantity):
            if duration.dimensions != b2.second.dimensions:
                raise TypeError("duration must have time dimensions if it's a Brian2 Quantity.")
            plot_duration_val = float(duration / time_unit_display)
        elif isinstance(duration, (int, float)):
            plot_duration_val = float(duration)
        else:
            raise TypeError("duration must be a Brian2 Quantity, a number, or None.")
    # --- FIM DA CORREÇÃO ---

    if len(plot_times_val) > 0 or plot_duration_val is not None:
        ax.plot(plot_times_val, plot_indices_val, '|', markersize=marker_size, color=marker_color)
        ax.set_xlabel(f"Time ({time_unit_display!s})")
        ax.set_ylabel(ylabel)
        ax.set_title(title)

        if plot_duration_val is not None:
            ax.set_xlim([0, plot_duration_val])
        elif len(plot_times_val) > 0:
            min_t, max_t = np.min(plot_times_val), np.max(plot_times_val)
            padding = 0.05 * (max_t - min_t) if (max_t - min_t) > 1e-9 else 0.05 * max_t
            if padding == 0 and max_t == 0 : padding = 1.0
            ax.set_xlim([max(0, min_t - padding), max_t + padding])
        else:
            ax.set_xlim([0, 1])

        if len(plot_indices_val) > 0:
            ax.set_ylim([np.min(plot_indices_val) - 0.5, np.max(plot_indices_val) + 0.5])
        else:
            ax.set_ylim([-0.5, 0.5])
    else:
        ax.set_xlabel(f"Time ({time_unit_display!s})")
        ax.set_ylabel(ylabel)
        ax.set_title(title + " (No data)")
        ax.set_xlim([0,1]); ax.set_ylim([-0.5, 0.5])

    ax.grid(True, linestyle=':', alpha=0.7)
    return ax


def plot_vm_traces(state_monitor, neuron_indices=None, time_unit_display=b2.ms, voltage_unit_display=b2.mV,
                   ax=None, title="Membrane Potential Traces",
                   xlabel=None, ylabel=None, legend_loc="best", alpha=0.8):
    """
    Plots membrane potential (Vm) traces for selected neurons from a StateMonitor.
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=(10, 4))

    # --- CORREÇÃO PARA LIDAR COM VariableView E Quantity PARA DADOS DE VOLTAGEM E TEMPO ---
    times_qty = state_monitor.t # This is a VariableView
    
    # Determine which variable to plot as Vm
    voltage_data_raw = None # This will be the VariableView or Quantity
    voltage_var_name_recorded = ""

    if hasattr(state_monitor, 'v'): # Common case where 'v' is directly an attribute
        voltage_data_raw = state_monitor.v
        voltage_var_name_recorded = 'v'
    elif state_monitor.variables: # Check the dict of recorded variables if 'v' is not a direct attr
        if 'v' in state_monitor.variables:
            voltage_data_raw = getattr(state_monitor, 'v') # It might still be an attribute
            voltage_var_name_recorded = 'v'
        else: # Fallback to the first recorded variable
            first_var_key = list(state_monitor.variables.keys())[0]
            voltage_data_raw = getattr(state_monitor, first_var_key)
            voltage_var_name_recorded = first_var_key
            print(f"Info: Plotting variable '{voltage_var_name_recorded}' as Vm from StateMonitor.")
    else:
        raise AttributeError("StateMonitor does not seem to have recorded 'v' or any other variables.")
    
    if voltage_data_raw is None:
        raise AttributeError("Could not extract voltage data from StateMonitor.")

    # Get numerical values in desired display units
    try:
        times_val = np.asarray(times_qty / time_unit_display)
    except Exception as e:
        raise TypeError(f"Could not process StateMonitor.t. Error: {e}")
    
    try:
        voltages_val = np.asarray(voltage_data_raw / voltage_unit_display)
    except Exception as e:
        raise TypeError(f"Could not process StateMonitor variable '{voltage_var_name_recorded}'. Error: {e}")
    # --- FIM DA CORREÇÃO ---

    num_actually_recorded_neurons = voltages_val.shape[0]
    indices_in_monitor_to_plot = []

    if neuron_indices is None:
        if num_actually_recorded_neurons <= 5:
            indices_in_monitor_to_plot = list(range(num_actually_recorded_neurons))
        elif num_actually_recorded_neurons > 0:
            raise ValueError(f"{num_actually_recorded_neurons} neurons recorded by StateMonitor. Please specify `neuron_indices`.")
    elif isinstance(neuron_indices, int):
        if not (0 <= neuron_indices < num_actually_recorded_neurons):
            raise ValueError(f"neuron_index {neuron_indices} out of bounds for monitor's data (0-{num_actually_recorded_neurons-1}).")
        indices_in_monitor_to_plot = [neuron_indices]
    elif isinstance(neuron_indices, (list, slice, np.ndarray)):
        # (Lógica para processar list/slice como antes)
        if isinstance(neuron_indices, slice):
            indices_in_monitor_to_plot = list(range(*neuron_indices.indices(num_actually_recorded_neurons)))
        else: 
            indices_in_monitor_to_plot = list(neuron_indices)
        for idx_mon in indices_in_monitor_to_plot:
            if not (0 <= idx_mon < num_actually_recorded_neurons):
                raise ValueError(f"Neuron index {idx_mon} out of bounds for monitor's data (0-{num_actually_recorded_neurons-1}).")
    else:
        raise TypeError("neuron_indices must be None, int, list, slice, or np.ndarray.")

    if not indices_in_monitor_to_plot and num_actually_recorded_neurons > 0:
        print("Warning: No specific neuron indices selected for plotting Vm.")
    elif num_actually_recorded_neurons == 0:
        print("Warning: StateMonitor recorded no data or no neurons.")

    for monitor_idx_to_plot in indices_in_monitor_to_plot:
        original_neuron_label = "" # (Lógica para rótulo como antes)
        if isinstance(state_monitor.record, (list, np.ndarray)): original_neuron_label = f"Neuron {state_monitor.record[monitor_idx_to_plot]}"
        elif isinstance(state_monitor.record, slice):
            start = state_monitor.record.start if state_monitor.record.start is not None else 0
            step = state_monitor.record.step if state_monitor.record.step is not None else 1
            original_neuron_label = f"Neuron {start + monitor_idx_to_plot * step}"
        elif state_monitor.record is True: original_neuron_label = f"Neuron {monitor_idx_to_plot}"
        else: original_neuron_label = f"Trace {monitor_idx_to_plot}"
        ax.plot(times_val, voltages_val[monitor_idx_to_plot, :], label=original_neuron_label, alpha=alpha)

    ax.set_title(title)
    ax.set_xlabel(xlabel if xlabel is not None else f"Time ({time_unit_display!s})")
    ax.set_ylabel(ylabel if ylabel is not None else f"Vm ({voltage_var_name_recorded} in {voltage_unit_display!s})") # More specific ylabel

    if legend_loc and len(indices_in_monitor_to_plot) > 0:
        ax.legend(loc=legend_loc, fontsize='small')

    ax.grid(True, linestyle=':', alpha=0.7)
    return ax